(* Routines for the management of XMS-memory from Turbo-Pascal programs	*)
(*									*)
(* author:	Hendrik Lbben	; date of begin:		010696	*)
(*				; date of last change:		050696	*)
(*				; time of development (hours):	2	*)
(* client:								*)
(* components:	Executables 0, Sources 1, Extra-documentation 0		*)
(* programbase:	XMSP.PAS (by M. Tischer, PC-Intern 3.0, DATA Becker)	*)
(*									*)
(*   C O M P U M A N   ---						*)
(*   If you exert changes on this software and then hand it on,		*)
(*   please keep this header.						*)
(*   Do not forget to add the information, what changes you made.	*)
(*	Berlin, June 1996						*)
(*	Hendrik Lbben, compuman@cs.tu-berlin.de			*)
(*						---   C O M P U M A N	*)


{ Attention:								}
{	Apart from the automatic check for an XMS driver at the start	}
{	of the program, no error-checks are performed by any of the XMS	}
{	routines. After each function call you should query XMSerr. Any	}
{	value different from zero means that the last operation caused	}
{	an error. A list of error-codes is provided at the start of the	}
{	interface-part of this unit.					}


unit XMS;



interface
const ERR_NOERR         =$00;
      ERR_NOTIMPLEMENTED=$80;
      ERR_VDISKFOUND    =$81;
      ERR_A20           =$82;	{ error on addressline A20		}
      ERR_GENERAL       =$8E;
      ERR_UNRECOVERABLE =$8F;
      ERR_A20STILLON    =$94;	{ addressline A20 still on		}
      ERR_OUTOMEMORY    =$A0;
      ERR_OUTOHANDLES   =$A1;
      ERR_INVALIDHANDLE =$A2;
      ERR_SHINVALID     =$A3;	{ source-handle invalid			}
      ERR_SOINVALID     =$A4;	{ source-offset invalid			}
      ERR_DHINVALID     =$A5;	{ destination-handle invalid		}
      ERR_DOINVALID     =$A6;	{ destination-offset invalid		}
      ERR_LENINVALID    =$A7;   { invalid length for move function	}
      ERR_OVERLAP       =$A8;	{ illegal overlap			}
      ERR_PARITY        =$A9;
var XMSerr:byte;

procedure XMSQueryFree(var GesFrei,MaxBl:integer);
function  XMSGetMem(LenKb:integer):integer;
procedure XMSFreeMem(Handle:integer);
procedure XMSCopy(FromHandle:integer;FromOffset:longint;
                  ToHandle:integer;ToOffset,LenW:longint);
{ seldomly used, therefore undocumented		}
{ try out or look up in PC-Intern (see above)	}
function  XMSLock(Handle:integer):longint;
procedure XMSUnLock(Handle:integer);
procedure XMSQueryInfo(hand:integer; var Lock,LenKB,FreeH:integer);
function  XMSrealloc(Handle,NeuLenKB:integer):boolean;



implementation
uses crt,dos;
			{ Registers needed for the XMScall-routine	}
type XMSregs=record AX,BX,DX,SI,DS:word; end;
var XMSptr:pointer;

{ initialization-routine for XMS-access					}
{ executed automatically upon start of any program using this unit	}
function XMSinit:boolean;
var R:registers;
begin
 R.AX:=$4300; intr($2F,R);		{ look for XMS-driver }
 if R.AL=$80 then begin			{ present ? }
  R.AX:=$4310; intr($2F,R);		{ yes -> query its address }
  XMSptr:=ptr(R.ES,R.BX); XMSerr:=ERR_NOERR; XMSinit:=TRUE;
 end else XMSinit:=FALSE;
end;

{$F-}
procedure XMScall(var XRegs:XMSregs);
begin
 inline($8C / $D9 /		{ MOV	CX,DS }
	$51 /			{ PUSH	CX }
	$C5 / $7E / $04 /	{ LDS	DI,[BP+04] ; [begin of routine+returnaddress] }
	$8B / $05 /		{ MOV	AX,[DI]    ;  = begin of Data }
	$8B / $5D / $02 /	{ MOV	BX,[DI+02] }
	$8B / $55 / $04 /	{ MOV	DX,[DI+04] }
	$8B / $75 / $06 /	{ MOV	SI,[DI+06] }
	$8E / $5D / $08 /	{ MOV	DS,[DI+08] }
	$8E / $C1 /		{ MOV	ES,CX }
	$26 /			{ ES: }
	$FF / $1E / XMSptr /	{ CALL	[XMSptr] }
	$8C / $D9 /		{ MOV	CX,DS }
	$C5 / $7E / $04 /	{ LDS	DI,[BP+04] }
	$89 / $05 /		{ MOV	[DI],AX }
	$89 / $5D / $02 /	{ MOV	[DI+02],BX }
	$89 / $55 / $04 /	{ MOV	[DI+04],DX }
	$89 / $75 / $06 /	{ MOV	[DI+06],SI }
	$89 / $4D / $08 /	{ MOV	[DI+08],CX }
	$1F);			{ POP	DS }
 if (XRegs.AX=0) and (XRegs.BX>=$80)
 then XMSerr:=Lo(XRegs.BX) else XMSerr:=0;
end;
{$F+}


(*function XMSQueryVer(var VerNr,iRevNr:word):boolean;
var Xr:XMSregs;
begin
 Xr.AX:=0; XmsCall(Xr);
 VerNr:= Hi(Xr.AX)*100 + (Lo(Xr.AX) SHR 4)*10 + (Lo(Xr.AX) AND $F);
 iRevNr:=Hi(Xr.BX)*100 + (Lo(Xr.BX) SHR 4)*10 + (Lo(Xr.BX) AND $F);
 XMSQueryVer:=(Xr.DX=1);
end;*)
(* var VerNr,RevNr:word; *)
(* writeln('XMS-version      : ',VerNr div 100,'.',VerNr mod 100); *)
(* writeln('internal revision: ',RevNr div 100,'.',RevNr mod 100); *)


{ returns the amount of free XMS (entire, largest block) in KB		}
procedure XMSQueryFree(var GesFrei,MaxBl:integer);
var Xr:XMSregs;
begin Xr.AX:=$800; XmsCall(Xr); GesFrei:=Xr.AX; MaxBl:=Xr.DX; end;

{ allocates the requested amount of XMS (in KB) and returns a handle	}
{ which is needed for any future access to the allocated XMS-area	}
function XMSGetMem(LenKb:integer):integer;
var Xr:XMSregs;
begin Xr.DX:=LenKB; Xr.AX:=$900; XmsCall(Xr); XMSGetMem:=Xr.DX; end;

{ releases the previously allocted XMS-area referred to by Handle	}
procedure XMSFreeMem(Handle:integer);
var Xr:XMSregs; begin Xr.AX:=$A00; Xr.DX:=Handle; XmsCall(Xr); end;

{ copies data between conventional RAM and the XMS			}
{ FromHandle/FromOffset specifies the source, ToHandle/ToOffset the	}
{ destination and LenW the amount vom words to be copied		}
{ an adress in convetional RAM is given by handle 0 and the complete	}
{ adress as offset (casted as longint)					}
{ an XMS-adress is given by the corresponding handle (see above) and	}
{ a simple linear offset into the XMS-area				}
procedure XMSCopy(FromHandle:integer;FromOffset:longint;
                  ToHandle:integer;ToOffset,LenW:longint);
type EMMS=record { an Extended-Memory-Move-structure }
           LenB:longint;
           SHandle:integer; SOffset:longint;
           DHandle:integer; DOffset:longint;
          end;
var Xr:XMSregs; Mi:EMMS;
begin
 with Mi do begin
  LenB:=LenW SHL 1;
  SHandle:=FromHandle; SOffset:=FromOffset;
  DHandle:=ToHandle;   DOffset:=ToOffset;
 end;
 Xr.AX:=$B00; Xr.SI:=Ofs(Mi); Xr.DS:=Seg(Mi); XMScall(Xr);
end;


function XMSLock(Handle:integer):longint;
var Xr:XMSregs;
begin
 Xr.AX:=$C00; Xr.DX:=Handle; XmsCall(Xr);
 XMSlock:=longint(Xr.DX) SHL 16+Xr.BX;
end;

procedure XMSUnLock(Handle:integer);
var Xr:XMSregs; begin Xr.AX:=$D00; Xr.DX:=Handle; XMScall(Xr); end;

procedure XMSQueryInfo(hand:integer; var Lock,LenKB,FreeH:integer);
var Xr:XMSregs;
begin
 Xr.AX:=$E00; Xr.DX:=hand; XmsCall(Xr);
 Lock:=Hi(Xr.BX); FreeH:=Lo(Xr.BX); LenKB:=Xr.DX;
end;

function XMSrealloc(Handle,NeuLenKB:integer):boolean;
var Xr:XMSregs;
begin
 Xr.AX:=$F00; Xr.DX:=Handle; Xr.BX:=NeuLenKB; XMScall(Xr);
 XMSrealloc:=(XMSerr=ERR_NOERR);
end;


begin
 if not XMSInit then
 begin writeln('XMS not accessible, program aborted !'); halt(1); end;
end.
